"""Initial revision

Revision ID: 48f9a7635310
Revises: 
Create Date: 2023-12-27 01:35:29.099926

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "48f9a7635310"
down_revision = None
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "suppgram_agents",
        sa.Column("id", sa.Integer(), autoincrement=True, nullable=False),
        sa.Column("deactivated", sa.Boolean(), nullable=False),
        sa.Column("telegram_user_id", sa.Integer(), nullable=True),
        sa.Column("telegram_first_name", sa.String(), nullable=True),
        sa.Column("telegram_last_name", sa.String(), nullable=True),
        sa.Column("telegram_username", sa.String(), nullable=True),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("telegram_user_id"),
    )
    op.create_table(
        "suppgram_customers",
        sa.Column("id", sa.Integer(), autoincrement=True, nullable=False),
        sa.Column("telegram_user_id", sa.Integer(), nullable=True),
        sa.Column("telegram_first_name", sa.String(), nullable=True),
        sa.Column("telegram_last_name", sa.String(), nullable=True),
        sa.Column("telegram_username", sa.String(), nullable=True),
        sa.Column("shell_uuid", sa.String(), nullable=True),
        sa.Column("pubnub_user_id", sa.String(), nullable=True),
        sa.Column("pubnub_channel_id", sa.String(), nullable=True),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("shell_uuid"),
        sa.UniqueConstraint("telegram_user_id"),
    )
    op.create_table(
        "suppgram_telegram_chats",
        sa.Column("telegram_chat_id", sa.Integer(), nullable=False),
        sa.Column("roles", sa.Integer(), nullable=False),
        sa.PrimaryKeyConstraint("telegram_chat_id"),
    )
    op.create_table(
        "suppgram_tags",
        sa.Column("id", sa.Integer(), autoincrement=True, nullable=False),
        sa.Column("name", sa.String(), nullable=False),
        sa.Column("created_at_utc", sa.DateTime(timezone=True), nullable=False),
        sa.Column("created_by_id", sa.Integer(), nullable=False),
        sa.ForeignKeyConstraint(
            ["created_by_id"],
            ["suppgram_agents.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("name"),
    )
    op.create_table(
        "suppgram_workplaces",
        sa.Column("id", sa.Integer(), autoincrement=True, nullable=False),
        sa.Column("agent_id", sa.Integer(), nullable=False),
        sa.Column("telegram_bot_id", sa.Integer(), nullable=False),
        sa.ForeignKeyConstraint(
            ["agent_id"],
            ["suppgram_agents.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_table(
        "suppgram_conversations",
        sa.Column("id", sa.Integer(), autoincrement=True, nullable=False),
        sa.Column("customer_id", sa.Integer(), nullable=False),
        sa.Column("assigned_workplace_id", sa.Integer(), nullable=True),
        sa.Column(
            "state",
            sa.Enum("NEW", "ASSIGNED", "RESOLVED", name="conversationstate"),
            nullable=False,
        ),
        sa.Column("customer_rating", sa.Integer(), nullable=True),
        sa.ForeignKeyConstraint(
            ["assigned_workplace_id"],
            ["suppgram_workplaces.id"],
        ),
        sa.ForeignKeyConstraint(
            ["customer_id"],
            ["suppgram_customers.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_table(
        "suppgram_conversation_messages",
        sa.Column("id", sa.Integer(), autoincrement=True, nullable=False),
        sa.Column("conversation_id", sa.Integer(), nullable=False),
        sa.Column(
            "kind",
            sa.Enum("FROM_CUSTOMER", "FROM_AGENT", "POSTPONED", "RESOLVED", name="messagekind"),
            nullable=False,
        ),
        sa.Column("time_utc", sa.DateTime(timezone=True), nullable=False),
        sa.Column("text", sa.String(), nullable=True),
        sa.ForeignKeyConstraint(
            ["conversation_id"],
            ["suppgram_conversations.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_table(
        "suppgram_conversation_tag_associations",
        sa.Column("conversation_id", sa.Integer(), nullable=False),
        sa.Column("tag_id", sa.Integer(), nullable=False),
        sa.ForeignKeyConstraint(
            ["conversation_id"],
            ["suppgram_conversations.id"],
        ),
        sa.ForeignKeyConstraint(
            ["tag_id"],
            ["suppgram_tags.id"],
        ),
        sa.PrimaryKeyConstraint("conversation_id", "tag_id"),
    )
    op.create_table(
        "suppgram_telegram_messages",
        sa.Column("id", sa.Integer(), autoincrement=True, nullable=False),
        sa.Column("chat_id", sa.Integer(), nullable=False),
        sa.Column("telegram_bot_id", sa.Integer(), nullable=False),
        sa.Column("telegram_message_id", sa.Integer(), nullable=False),
        sa.Column(
            "kind",
            sa.Enum(
                "NEW_CONVERSATION_NOTIFICATION",
                "RATE_CONVERSATION",
                "NUDGE_TO_START_BOT_NOTIFICATION",
                "CUSTOMER_MESSAGE_HISTORY",
                name="telegrammessagekind",
            ),
            nullable=False,
        ),
        sa.Column("agent_id", sa.Integer(), nullable=True),
        sa.Column("customer_id", sa.Integer(), nullable=True),
        sa.Column("conversation_id", sa.Integer(), nullable=True),
        sa.Column("telegram_bot_username", sa.String(), nullable=True),
        sa.ForeignKeyConstraint(
            ["agent_id"],
            ["suppgram_agents.id"],
        ),
        sa.ForeignKeyConstraint(
            ["chat_id"],
            ["suppgram_telegram_chats.telegram_chat_id"],
        ),
        sa.ForeignKeyConstraint(
            ["conversation_id"],
            ["suppgram_conversations.id"],
        ),
        sa.ForeignKeyConstraint(
            ["customer_id"],
            ["suppgram_customers.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
    )
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table("suppgram_telegram_messages")
    op.drop_table("suppgram_conversation_tag_associations")
    op.drop_table("suppgram_conversation_messages")
    op.drop_table("suppgram_conversations")
    op.drop_table("suppgram_workplaces")
    op.drop_table("suppgram_tags")
    op.drop_table("suppgram_telegram_chats")
    op.drop_table("suppgram_customers")
    op.drop_table("suppgram_agents")
    # ### end Alembic commands ###



from typing import TypeVar, List, Iterator, overload, SupportsIndex

from suppgram.errors import DataNotFetched

T = TypeVar("T")


class UnavailableList(List[T]):
    def __len__(self) -> int:
        raise DataNotFetched(
            "data for this list have not been fetched from the database"
        )

    def __iter__(self) -> Iterator[T]:
        raise DataNotFetched(
            "data for this list have not been fetched from the database"
        )

    @overload
    def __getitem__(self, index: SupportsIndex) -> T:
        raise DataNotFetched(
            "data for this list have not been fetched from the database"
        )

    @overload
    def __getitem__(self, sl: slice) -> List[T]:
        raise DataNotFetched(
            "data for this list have not been fetched from the database"
        )

    def __getitem__(self, *args, **kwargs):
        raise DataNotFetched(
            "data for this list have not been fetched from the database"
        )

    def __setitem__(self, *args, **kwargs) -> None:
        raise DataNotFetched(
            "data for this list have not been fetched from the database"
        )

    def __delitem__(self, *args, **kwargs) -> None:
        raise DataNotFetched(
            "data for this list have not been fetched from the database"
        )



from typing import List, Mapping, cast

from suppgram.backend import WorkplaceManager
from suppgram.entities import Agent, Workplace, WorkplaceIdentification
from suppgram.frontends.telegram.app_manager import TelegramAppManager
from suppgram.helpers import flat_gather


class TelegramWorkplaceManager(WorkplaceManager):
    def __init__(self, tokens: List[str], app_manager: TelegramAppManager):
        self._tokens = tokens
        self._app_manager = app_manager
        self._bot_id_to_index: Mapping[int, int] = {}

    async def initialize(self):
        await super().initialize()
        await flat_gather(self._app_manager.get_app(token).initialize() for token in self._tokens)

    def create_missing_workplaces(
        self, agent: Agent, existing_workplaces: List[Workplace]
    ) -> List[WorkplaceIdentification]:
        bot_ids = self._get_bot_id_to_index()
        present_bot_ids = {workplace.telegram_bot_id for workplace in existing_workplaces}
        missing_bot_ids = set(bot_ids.keys()) - present_bot_ids
        return [
            WorkplaceIdentification(telegram_user_id=agent.telegram_user_id, telegram_bot_id=bot_id)
            for bot_id in missing_bot_ids
        ]

    def filter_and_rank_available_workplaces(self, workplaces: List[Workplace]) -> List[Workplace]:
        bot_ids = self._get_bot_id_to_index()
        available_workplaces = [
            workplace for workplace in workplaces if workplace.telegram_bot_id in bot_ids
        ]
        available_workplaces.sort(key=lambda w: bot_ids[cast(int, w.telegram_bot_id)])
        return available_workplaces

    def _get_bot_id_to_index(self) -> Mapping[int, int]:
        if not self._bot_id_to_index:
            self._bot_id_to_index = {
                self._app_manager.get_app(token).bot.id: idx
                for idx, token in enumerate(self._tokens)
            }
        return self._bot_id_to_index



import re

from suppgram.emoji import EMOJI_SEQUENCE


def test_emoji():
    regexp = re.compile(f"{EMOJI_SEQUENCE}+")
    assert not regexp.match("x")
    assert regexp.match("😅")
    assert regexp.match("👷🏿")
    # assert not regexp.match("🏿")
    assert regexp.match("‼️")



import os
from dataclasses import dataclass
from typing import Optional

from pubnub.pnconfiguration import PNConfiguration
from pubnub.pubnub_asyncio import PubNubAsyncio

from suppgram.frontends.pubnub.errors import MissingCredentials


@dataclass(frozen=True)
class Configuration:
    pn_config: PNConfiguration
    pn_token: Optional[str]

    def instantiate_async(self) -> PubNubAsyncio:
        pubnub = PubNubAsyncio(self.pn_config)
        if self.pn_token:
            pubnub.set_token(self.pn_token)
        return pubnub


def make_pubnub_configuration(pubnub_user_id: str) -> Configuration:
    pn_config = PNConfiguration()
    pn_config.subscribe_key = _get_env_variable("PUBNUB_SUBSCRIBE_KEY")
    pn_config.publish_key = _get_env_variable("PUBNUB_PUBLISH_KEY")
    pn_config.secret_key = os.environ.get("PUBNUB_SECRET_KEY")
    pn_config.uuid = pubnub_user_id
    pn_token = os.environ.get("PUBNUB_TOKEN")

    if not pn_config.secret_key and not pn_token:
        raise MissingCredentials(
            "both PUBNUB_SECRET_KEY and PUBNUB_TOKEN environment variables are not set — "
            "the app won't have permissions to do anything.\n\n"
            "Provide PubNub secret key if you trust the app; otherwise, "
            "create access token with necessary permissions (User.get for the support user, "
            "Channel.read, Channel.write and Channel.join for the support channels group)."
        )

    return Configuration(pn_config=pn_config, pn_token=pn_token)


def _get_env_variable(name: str) -> str:
    try:
        return os.environ[name]
    except KeyError as exc:
        raise MissingCredentials(
            f"required environment variable {name} is not set"
        ) from exc



from datetime import datetime, timezone
from typing import List, Any, Optional

from suppgram.backend import Backend as BackendInterface, WorkplaceManager
from suppgram.entities import (
    ConversationEvent,
    NewMessageForCustomerEvent,
    NewUnassignedMessageFromCustomerEvent,
    NewMessageForAgentEvent,
    AgentIdentification,
    Agent,
    AgentDiff,
    CustomerIdentification,
    Conversation,
    WorkplaceIdentification,
    Workplace,
    Message,
    ConversationState,
    ConversationDiff,
    SetNone,
    MessageKind,
    CustomerDiff,
    Tag,
    ConversationTagEvent,
    Customer,
    TagEvent,
    Event,
    EventKind,
)
from suppgram.errors import PermissionDenied, AgentDeactivated
from suppgram.helpers import flat_gather
from suppgram.observer import LocalObservable
from suppgram.storage import Storage
from suppgram.texts.en import EnglishTextProvider
from suppgram.texts.interface import TextProvider


class LocalBackend(BackendInterface):
    def __init__(
        self,
        storage: Storage,
        workplace_managers: List[WorkplaceManager],
        texts: TextProvider = EnglishTextProvider(),
    ):
        self._storage = storage
        self._workplace_managers = workplace_managers
        self._texts = texts

        self.on_new_conversation = LocalObservable[ConversationEvent]()
        self.on_conversation_assignment = LocalObservable[ConversationEvent]()
        self.on_conversation_resolution = LocalObservable[ConversationEvent]()
        self.on_conversation_tag_added = LocalObservable[ConversationTagEvent]()
        self.on_conversation_tag_removed = LocalObservable[ConversationTagEvent]()
        self.on_conversation_rated = LocalObservable[ConversationEvent]()
        self.on_new_message_for_customer = LocalObservable[NewMessageForCustomerEvent]()
        self.on_new_unassigned_message_from_customer = LocalObservable[
            NewUnassignedMessageFromCustomerEvent
        ]()
        self.on_new_message_for_agent = LocalObservable[NewMessageForAgentEvent]()
        self.on_tag_created = LocalObservable[TagEvent]()

    async def create_or_update_agent(
        self, identification: AgentIdentification, diff: Optional[AgentDiff] = None
    ) -> Agent:
        agent = await self._storage.create_or_update_agent(identification, diff)
        existing_workplaces = await self._storage.get_agent_workplaces(agent)
        await self._create_all_missing_workplaces(agent, existing_workplaces)
        return agent

    async def identify_agent(self, identification: AgentIdentification) -> Agent:
        return await self._storage.get_agent(identification)

    async def update_agent(self, identification: AgentIdentification, diff: AgentDiff) -> Agent:
        if diff.deactivated:
            raise ValueError("can't deactivate agent via diff, use `deactivate_agent()`")
        return await self._storage.update_agent(identification, diff)

    async def deactivate_agent(self, agent: Agent):
        # TODO transaction?..
        conversations = await self._storage.find_agent_conversations(agent)
        await flat_gather(self.postpone_conversation(agent, conv) for conv in conversations)
        await self._storage.update_agent(agent.identification, AgentDiff(deactivated=True))

    async def create_or_update_customer(
        self, identification: CustomerIdentification, diff: Optional[CustomerDiff] = None
    ) -> Customer:
        return await self._storage.create_or_update_customer(identification, diff)

    async def identify_customer_conversation(
        self, identification: CustomerIdentification
    ) -> Conversation:
        customer = await self._storage.create_or_update_customer(identification)
        conversation = await self._storage.get_or_create_conversation(customer)
        return conversation

    async def identify_workplace(self, identification: WorkplaceIdentification) -> Workplace:
        return await self._storage.get_or_create_workplace(identification)

    async def create_tag(self, name: str, created_by: Agent) -> Tag:
        if created_by.deactivated:
            raise AgentDeactivated(created_by.identification)

        tag = await self._storage.create_tag(name=name, created_by=created_by)
        await self.on_tag_created.trigger(TagEvent(tag=tag))
        return tag

    async def get_all_tags(self) -> List[Tag]:
        return await self._storage.find_all_tags()

    async def identify_agent_conversation(
        self, identification: WorkplaceIdentification
    ) -> Conversation:
        return await self._storage.get_agent_conversation(identification)

    async def process_message(self, conversation: Conversation, message: Message):
        if message.kind == MessageKind.FROM_CUSTOMER:
            await self._process_message_from_user(conversation, message)
        elif message.kind == MessageKind.FROM_AGENT:
            await self._process_message_from_agent(conversation, message)
        else:
            await self._process_internal_message(conversation, message)

    async def _process_message_from_user(self, conversation: Conversation, message: Message):
        await self._storage.save_message(conversation, message)
        conversation.messages.append(message)
        if len(conversation.messages) == 1:
            await self.on_new_conversation.trigger(ConversationEvent(conversation))
            await self._storage.save_event(
                Event(
                    kind=EventKind.CONVERSATION_STARTED,
                    time_utc=message.time_utc,
                    conversation_id=conversation.id,
                    customer_id=conversation.customer.id,
                )
            )
        if conversation.assigned_agent and conversation.assigned_workplace:
            await self.on_new_message_for_agent.trigger(
                NewMessageForAgentEvent(
                    agent=conversation.assigned_agent,
                    workplace=conversation.assigned_workplace,
                    message=message,
                )
            )
        else:
            await self.on_new_unassigned_message_from_customer.trigger(
                NewUnassignedMessageFromCustomerEvent(message=message, conversation=conversation)
            )
        await self._storage.save_event(
            Event(
                kind=EventKind.MESSAGE_SENT,
                time_utc=message.time_utc,
                conversation_id=conversation.id,
                customer_id=conversation.customer.id,
                message_kind=message.kind,
                message_media_kind=message.media_kind,
            )
        )

    async def _process_message_from_agent(self, conversation: Conversation, message: Message):
        if conversation.assigned_agent and conversation.assigned_agent.deactivated:
            raise AgentDeactivated(conversation.assigned_agent.identification)

        await self._storage.save_message(conversation, message)
        await self.on_new_message_for_customer.trigger(
            NewMessageForCustomerEvent(
                conversation=conversation,
                message=message,
            )
        )
        await self._storage.save_event(
            Event(
                kind=EventKind.MESSAGE_SENT,
                time_utc=message.time_utc,
                conversation_id=conversation.id,
                customer_id=conversation.customer.id,
                message_kind=message.kind,
                message_media_kind=message.media_kind,
            )
        )

    async def _process_internal_message(self, conversation: Conversation, message: Message):
        await self._storage.save_message(conversation, message)
        await self.on_new_message_for_customer.trigger(
            NewMessageForCustomerEvent(
                conversation=conversation,
                message=message,
            )
        )
        if conversation.assigned_agent and conversation.assigned_workplace:
            await self.on_new_message_for_agent.trigger(
                NewMessageForAgentEvent(
                    agent=conversation.assigned_agent,
                    workplace=conversation.assigned_workplace,
                    message=message,
                )
            )

    async def assign_agent(self, assigner: Agent, assignee: Agent, conversation_id: Any):
        if assigner.deactivated:
            raise AgentDeactivated(assigner.identification)
        if assignee.deactivated:
            raise AgentDeactivated(assignee.identification)

        workplace = await self._choose_workplace(assignee)
        await self._storage.update_conversation(
            conversation_id,
            ConversationDiff(state=ConversationState.ASSIGNED, assigned_workplace_id=workplace.id),
            unassigned_only=True,
        )
        conversation = await self._storage.get_agent_conversation(workplace.identification)
        await self.on_conversation_assignment.trigger(ConversationEvent(conversation=conversation))
        await self._storage.save_event(
            Event(
                kind=EventKind.AGENT_ASSIGNED,
                agent_id=assignee.id,
                conversation_id=conversation.id,
                customer_id=conversation.customer.id,
            )
        )

    async def get_conversations(
        self, conversation_ids: List[Any], with_messages: bool = False
    ) -> List[Conversation]:
        return await self._storage.find_conversations_by_ids(
            conversation_ids, with_messages=with_messages
        )

    async def get_customer_conversations(self, customer: Customer) -> List[Conversation]:
        return await self._storage.find_customer_conversations(customer, with_messages=True)

    async def add_tag_to_conversation(self, conversation: Conversation, tag: Tag):
        await self._storage.update_conversation(conversation.id, ConversationDiff(added_tags=[tag]))
        conversation = await self.get_conversation(conversation.id)
        await self.on_conversation_tag_added.trigger(
            ConversationTagEvent(conversation=conversation, tag=tag)
        )
        await self._storage.save_event(
            Event(
                kind=EventKind.CONVERSATION_TAG_ADDED,
                conversation_id=conversation.id,
                customer_id=conversation.customer.id,
                tag_id=tag.id,
            )
        )

    async def remove_tag_from_conversation(self, conversation: Conversation, tag: Tag):
        await self._storage.update_conversation(
            conversation.id, ConversationDiff(removed_tags=[tag])
        )
        conversation = await self.get_conversation(conversation.id)
        await self.on_conversation_tag_removed.trigger(
            ConversationTagEvent(conversation=conversation, tag=tag)
        )
        await self._storage.save_event(
            Event(
                kind=EventKind.CONVERSATION_TAG_REMOVED,
                conversation_id=conversation.id,
                customer_id=conversation.customer.id,
                tag_id=tag.id,
            )
        )

    async def rate_conversation(self, conversation: Conversation, rating: int):
        await self._storage.update_conversation(
            conversation.id, ConversationDiff(customer_rating=rating)
        )
        conversation = await self.get_conversation(conversation.id)
        await self.on_conversation_rated.trigger(ConversationEvent(conversation=conversation))
        await self._storage.save_event(
            Event(
                kind=EventKind.CONVERSATION_RATED,
                conversation_id=conversation.id,
                customer_id=conversation.customer.id,
            )
        )

    async def postpone_conversation(self, postponer: Agent, conversation: Conversation):
        if postponer != conversation.assigned_agent:
            raise PermissionDenied("not allowed to postpone conversations of other agents")

        await self.process_message(
            conversation,
            Message(kind=MessageKind.POSTPONED, time_utc=datetime.now(timezone.utc)),
        )
        await self._storage.update_conversation(
            conversation.id,
            ConversationDiff(state=ConversationState.NEW, assigned_workplace_id=SetNone),
        )
        conversation = Conversation(  # TODO just fetch?..
            id=conversation.id,
            state=ConversationState.NEW,
            customer=conversation.customer,
            messages=conversation.messages,
            tags=conversation.tags,
        )
        await self.on_new_conversation.trigger(ConversationEvent(conversation=conversation))
        await self._storage.save_event(
            Event(
                kind=EventKind.CONVERSATION_POSTPONED,
                agent_id=postponer.id,
                conversation_id=conversation.id,
                customer_id=conversation.customer.id,
            )
        )

    async def resolve_conversation(self, resolver: Agent, conversation: Conversation):
        if resolver != conversation.assigned_agent:
            raise PermissionDenied("not allowed to resolve conversations of other agents")
        if resolver.deactivated:
            await self.postpone_conversation(resolver, conversation)
            raise AgentDeactivated(resolver.identification)

        # TODO processing message and updating conversation should be in a single transaction
        await self.process_message(
            conversation,
            Message(kind=MessageKind.RESOLVED, time_utc=datetime.now(timezone.utc)),
        )
        await self._storage.update_conversation(
            conversation.id,
            ConversationDiff(state=ConversationState.RESOLVED, assigned_workplace_id=SetNone),
        )
        conversation = Conversation(  # TODO just fetch?..
            id=conversation.id,
            state=ConversationState.RESOLVED,
            customer=conversation.customer,
            messages=conversation.messages,
            tags=conversation.tags,
        )
        await self.on_conversation_resolution.trigger(ConversationEvent(conversation=conversation))
        await self._storage.save_event(
            Event(
                kind=EventKind.CONVERSATION_POSTPONED,
                agent_id=resolver.id,
                conversation_id=conversation.id,
                customer_id=conversation.customer.id,
            )
        )

    async def _choose_workplace(self, agent: Agent) -> Workplace:
        existing_workplaces = await self._storage.get_agent_workplaces(agent)
        extra_workplaces = await self._create_all_missing_workplaces(agent, existing_workplaces)
        all_workplaces = existing_workplaces + extra_workplaces
        available_workplaces = await self._filter_all_available_workplaces(agent, all_workplaces)
        return available_workplaces[0]  # TODO handle index out of range

    async def _create_all_missing_workplaces(
        self, agent: Agent, existing_workplaces: List[Workplace]
    ) -> List[Workplace]:
        extra_workplaces: List[Workplace] = []
        for manager in self._workplace_managers:
            missing_workplace_identifications = manager.create_missing_workplaces(
                agent, existing_workplaces
            )
            extra_workplaces.extend(
                await flat_gather(
                    self._storage.get_or_create_workplace(  # TODO session/batch
                        workplace_identification
                    )
                    for workplace_identification in missing_workplace_identifications
                )
            )
        return extra_workplaces

    async def _filter_all_available_workplaces(
        self, agent: Agent, all_workplaces: List[Workplace]
    ) -> List[Workplace]:
        available_workplaces: List[Workplace] = []
        for manager in self._workplace_managers:
            available_workplaces.extend(
                manager.filter_and_rank_available_workplaces(all_workplaces)
            )
        agent_conversations = await self._storage.find_agent_conversations(agent)
        busy_workplace_ids = {
            conv.assigned_workplace.id
            for conv in agent_conversations
            if conv.assigned_workplace  # it can't be None, but for the sake of typing let's check
        }
        free_workplaces = [
            workplace
            for workplace in available_workplaces
            if workplace.id not in busy_workplace_ids
        ]
        return free_workplaces



import logging

from suppgram.entities import (
    Conversation,
    Message,
    MessageKind,
    Agent,
)
from suppgram.texts.interface import TextProvider, Text, Format

logger = logging.getLogger(__name__)


class EnglishTextProvider(TextProvider):
    telegram_customer_unsupported_message_content = (
        "😞 Sorry, this kind of content is not supported right now. "
        "Support agent will not see this message."
    )
    telegram_agent_unsupported_message_content = (
        "😞 Sorry, this kind of content is not supported right now. "
        "Customer will not see this message."
    )

    telegram_customer_start_message = "👋 Welcome to support service! Please describe your problem."
    telegram_customer_conversation_resolved_message_placeholder = (
        "✅ Conversation was marked as resolved. "
        "You can always start a new conversation by writing to this chat!"
    )
    telegram_customer_conversation_resolved_message = (
        telegram_customer_conversation_resolved_message_placeholder + "\n\n"
        "⭐️ Please rate your experience with the support agent:"
    )

    def compose_customer_conversation_resolved_message(self, rating: int) -> str:
        return (
            self.telegram_customer_conversation_resolved_message_placeholder
            + "\n\n⭐️ How you rated this conversation: "
            + self.format_rating(rating)
        )

    telegram_manager_start_message = "🛠️ Welcome to the support admin bot!"
    telegram_manager_permission_denied_message = (
        "🚫 You don't have permission to access manager functionality."
    )
    telegram_agents_command_description = "Make all members of this group support agents."
    telegram_agents_command_success_message = (
        "I will now consider all members of this group support agents."
    )
    telegram_send_new_conversations_command_description = (
        "Send notifications about new conversations to this group."
    )
    telegram_send_new_conversations_command_success_message = (
        "I will now send notifications about new conversations to this group."
    )

    telegram_report_command_description = "Export analytical report"
    telegram_report_message_placeholder = "⏳ Computing report... {progress}"
    telegram_report_message = (
        "📊 Analytical report\n\n"
        "Average time to first response: {report.average_start_to_first_response_time_min:.1f} min\n"
        "Average resolution time: {report.average_start_to_resolution_time_min:.1f} min\n"
        "Average customer satisfaction: {report.average_customer_rating:.1f}"
    )

    telegram_create_tag_command_description = "Create new tag to label conversations with."
    telegram_create_tag_permission_denied_message = (
        "🚫 You don't have permission to create new tags."
    )
    telegram_create_tag_usage_message = (
        "🧑‍🏫 Please specify new tag name after command:\n\n    /create_tag urgent"
    )
    telegram_tag_successfully_created_message = "✅ New tag has been successfully created."
    telegram_tag_already_exists_message = "⚠️ Tag with this name already exists!"

    telegram_agent_start_message = "👷 Welcome to the support agent bot!"
    telegram_agent_permission_denied_message = (
        "🚫 You don't have permission to access support agent functionality."
    )
    telegram_workplace_is_not_assigned_message = (
        "📭 This chat is not assigned to any ongoing conversation with a customer right now."
    )
    telegram_resolve_command_description = (
        "Mark conversation resolved and stop messaging with the customer."
    )
    telegram_postpone_command_description = (
        "Return the conversation to NEW status and stop messaging with the customer."
    )
    telegram_agent_conversation_resolved_message = (
        "✅ Conversation was marked as resolved. This chat is no longer assigned to a customer."
    )
    telegram_new_conversation_notification_placeholder = "❗️ New conversation!"

    customer_profile_header = "👤 Customer: {customer}"
    customer_profile_anonymous = "anonymous"
    customer_profile_contacts = "📒 Contacts: {contacts}"
    customer_rating_footer = "🎖 Rating: {rating}"

    def compose_conversation_notification_header(self, conversation: Conversation) -> str:
        emoji = self.CONVERSATION_STATE_TO_EMOJI.get(conversation.state, "")
        return f"{emoji} Conversation in status #{conversation.state.upper()}"

    conversation_notification_assigned_to = "Assigned to {agent}"

    def compose_nudge_to_start_bot_notification(
        self, agent: Agent, telegram_bot_username: str
    ) -> Text:
        if agent.telegram_user_id is None:
            raise RuntimeError("agent without Telegram account can't /start a bot!")
        agent_ref = self._format_telegram_mention(
            telegram_user_id=agent.telegram_user_id,
            telegram_first_name=agent.telegram_first_name,
            telegram_last_name=None,  # less formal
            telegram_username=agent.telegram_username,
            format_=Format.TELEGRAM_MARKDOWN,
        )
        return Text(
            text=f"⚠️ {agent_ref}, please open @{telegram_bot_username} and "
            f"hit Start/Restart button to be able to communicate with the customer\\.",
            format=Format.TELEGRAM_MARKDOWN,
        )

    telegram_assign_to_me_button_text = "Assign to me"

    message_history_title = "🗂️ Message history\n"

    def format_history_message(self, message: Message) -> str:
        if message.kind == MessageKind.FROM_CUSTOMER:
            return f"👤 Customer: {self.format_text_or_media(message)}"
        if message.kind == MessageKind.FROM_AGENT:
            return f"🧑‍💼 Agent: {self.format_text_or_media(message)}"
        if message.kind == MessageKind.POSTPONED:
            return "⏳ Conversation was postponed."
        if message.kind == MessageKind.RESOLVED:
            return "✅ Conversation was resolved."
        logger.warning(f"Unsupported message kind: {message.kind.value!r}")
        return str(message.kind.value)

    def format_text_or_media(self, message: Message) -> str:
        if message.text:
            return message.text
        if message.image:
            return "[🖼️image]"
        return f"[{message.media_kind}]"



import abc
from typing import Any, List, Optional

from suppgram.entities import (
    ConversationEvent,
    NewMessageForCustomerEvent,
    NewUnassignedMessageFromCustomerEvent,
    NewMessageForAgentEvent,
    AgentIdentification,
    Agent,
    AgentDiff,
    CustomerIdentification,
    Conversation,
    WorkplaceIdentification,
    Workplace,
    Message,
    CustomerDiff,
    Tag,
    ConversationTagEvent,
    Customer,
    TagEvent,
)
from suppgram.observer import Observable


class Backend(abc.ABC):
    on_new_conversation: Observable[ConversationEvent]
    on_conversation_assignment: Observable[ConversationEvent]
    on_conversation_resolution: Observable[ConversationEvent]
    on_conversation_tag_added: Observable[ConversationTagEvent]
    on_conversation_tag_removed: Observable[ConversationTagEvent]
    on_conversation_rated: Observable[ConversationEvent]
    on_new_message_for_customer: Observable[NewMessageForCustomerEvent]
    on_new_unassigned_message_from_customer: Observable[NewUnassignedMessageFromCustomerEvent]
    on_new_message_for_agent: Observable[NewMessageForAgentEvent]
    on_tag_created: Observable[TagEvent]

    @abc.abstractmethod
    async def create_or_update_agent(
        self, identification: AgentIdentification, diff: Optional[AgentDiff] = None
    ) -> Agent:
        pass

    @abc.abstractmethod
    async def identify_agent(self, identification: AgentIdentification) -> Agent:
        pass

    @abc.abstractmethod
    async def update_agent(self, identification: AgentIdentification, diff: AgentDiff) -> Agent:
        pass

    @abc.abstractmethod
    async def deactivate_agent(self, agent: Agent):
        pass

    @abc.abstractmethod
    async def create_or_update_customer(
        self, identification: CustomerIdentification, diff: Optional[CustomerDiff] = None
    ) -> Customer:
        pass

    @abc.abstractmethod
    async def identify_customer_conversation(
        self, identification: CustomerIdentification
    ) -> Conversation:
        pass

    @abc.abstractmethod
    async def identify_workplace(self, identification: WorkplaceIdentification) -> Workplace:
        pass

    @abc.abstractmethod
    async def create_tag(self, name: str, created_by: Agent) -> Tag:
        pass

    @abc.abstractmethod
    async def get_all_tags(self) -> List[Tag]:
        pass

    @abc.abstractmethod
    async def identify_agent_conversation(
        self, identification: WorkplaceIdentification
    ) -> Conversation:
        pass

    @abc.abstractmethod
    async def process_message(self, conversation: Conversation, message: Message):
        pass

    @abc.abstractmethod
    async def assign_agent(self, assigner: Agent, assignee: Agent, conversation_id: Any):
        pass

    async def get_conversation(self, conversation_id: Any) -> Conversation:
        return (await self.get_conversations([conversation_id], with_messages=True))[0]

    @abc.abstractmethod
    async def get_conversations(
        self, conversation_ids: List[Any], with_messages: bool = False
    ) -> List[Conversation]:
        pass

    @abc.abstractmethod
    async def get_customer_conversations(self, customer: Customer) -> List[Conversation]:
        pass

    @abc.abstractmethod
    async def add_tag_to_conversation(self, conversation: Conversation, tag: Tag):
        pass

    @abc.abstractmethod
    async def remove_tag_from_conversation(self, conversation: Conversation, tag: Tag):
        pass

    @abc.abstractmethod
    async def rate_conversation(self, conversation: Conversation, rating: int):
        pass

    @abc.abstractmethod
    async def postpone_conversation(self, resolver: Agent, conversation: Conversation):
        pass

    @abc.abstractmethod
    async def resolve_conversation(self, resolver: Agent, conversation: Conversation):
        pass


class WorkplaceManager(abc.ABC):
    async def initialize(self):
        pass

    @abc.abstractmethod
    def create_missing_workplaces(
        self, agent: Agent, existing_workplaces: List[Workplace]
    ) -> List[WorkplaceIdentification]:
        pass

    def filter_and_rank_available_workplaces(self, workplaces: List[Workplace]) -> List[Workplace]:
        return workplaces



import functools
import json
from itertools import zip_longest
from typing import Callable, TypeVar, Coroutine, List, Any, Iterable, Optional

from telegram import Update, InlineKeyboardButton, InlineKeyboardMarkup
from telegram.error import TelegramError
from telegram.ext import ContextTypes

from suppgram.frontends.telegram.callback_actions import CallbackActionKind
from suppgram.frontends.telegram.storage import TelegramMessage

T = TypeVar("T")


def send_text_answer(
    handler: Callable[[T, Update, ContextTypes.DEFAULT_TYPE], Coroutine[Any, Any, str]]
) -> Callable[[T, Update, ContextTypes.DEFAULT_TYPE], Coroutine[None, None, None]]:
    @functools.wraps(handler)
    async def wrapped_handler(self: T, update: Update, context: ContextTypes.DEFAULT_TYPE):
        assert update.effective_chat
        answer = await handler(self, update, context)
        await context.bot.send_message(update.effective_chat.id, answer)

    return wrapped_handler


def arrange_buttons(
    buttons: List[InlineKeyboardButton],
) -> List[List[InlineKeyboardButton]]:
    it = iter(buttons)
    return [[b1] if b2 is None else [b1, b2] for b1, b2 in zip_longest(it, it)]


def is_chat_not_found(exc: TelegramError) -> bool:
    return "Chat not found" in str(exc)


def is_blocked_by_user(exc: TelegramError) -> bool:
    return "bot was blocked by the user" in str(exc)


def encode_callback_data(doc: Any) -> str:
    return json.dumps(doc, separators=(",", ":"))


def decode_callback_data(data: str) -> Any:
    return json.loads(data)


def paginate_texts(
    prefix: str,
    texts: Iterable[str],
    suffix: str,
    delimiter: str = "\n",
    max_page_lines: int = 4000,
    max_page_chars: int = 4000,
) -> Iterable[str]:
    prefix_newlines = prefix.count("\n")
    delimiter_newlines = delimiter.count("\n")
    parts, parts_length, parts_newlines = [prefix], len(prefix), prefix_newlines
    suffix_newlines = suffix.count("\n")
    for text in texts:
        text_newlines = text.count("\n")
        page_length = parts_length + len(delimiter) + len(text) + len(delimiter) + len(suffix)
        page_newlines = (
            parts_newlines
            + delimiter_newlines
            + text_newlines
            + delimiter_newlines
            + suffix_newlines
        )
        if page_length > max_page_chars or page_newlines >= max_page_lines:
            parts.append(suffix)
            yield "".join(parts)
            parts, parts_length, parts_newlines = [prefix], len(prefix), prefix_newlines
        if parts_length > 0:
            parts.append(delimiter)
            parts_length += len(delimiter)
            parts_newlines += delimiter_newlines
        parts.append(text)
        parts_length += len(text)
        parts_newlines += text_newlines
    if parts_length > 0:
        parts.append(suffix)
        yield "".join(parts)


def make_pagination_keyboard(
    message: TelegramMessage, number_of_pages: int, current_page_idx: int
) -> Optional[InlineKeyboardMarkup]:
    if number_of_pages <= 1:
        return None

    if number_of_pages <= 5:
        min_page_idx, max_page_idx = 0, number_of_pages
    elif current_page_idx < 2:
        min_page_idx, max_page_idx = 0, 5
    elif current_page_idx + 3 >= number_of_pages:
        min_page_idx, max_page_idx = number_of_pages - 5, number_of_pages
    else:
        min_page_idx, max_page_idx = current_page_idx - 2, current_page_idx + 3

    buttons: List[InlineKeyboardButton] = []
    for page_idx in range(min_page_idx, max_page_idx):
        text = str(page_idx + 1)
        if page_idx == current_page_idx:
            buttons.append(InlineKeyboardButton(text=f"〈{text}〉", callback_data="{}"))
            continue
        if 0 < min_page_idx == page_idx:
            text = "«"
        if page_idx + 1 == max_page_idx < number_of_pages:
            text = "»"
        buttons.append(_make_pagination_button(message, text, page_idx))

    return InlineKeyboardMarkup(inline_keyboard=[buttons])


def _make_pagination_button(
    message: TelegramMessage, text: str, page_idx: int
) -> InlineKeyboardButton:
    return InlineKeyboardButton(
        text=text,
        callback_data=encode_callback_data(
            {
                "a": CallbackActionKind.PAGE,
                "c": message.chat.telegram_chat_id,
                "m": message.telegram_message_id,
                "p": page_idx,
            },
        ),
    )



import logging
from typing import cast, List

from telegram import Update, Bot, InlineKeyboardMarkup, InlineKeyboardButton
from telegram.ext import (
    CommandHandler,
    ContextTypes,
    MessageHandler,
    CallbackQueryHandler,
    BaseHandler,
)
from telegram.ext.filters import TEXT, ChatType, PHOTO

from suppgram.backend import Backend
from suppgram.entities import (
    MessageKind,
    Message,
    NewMessageForCustomerEvent,
    Conversation,
    ConversationEvent,
)
from suppgram.frontend import (
    CustomerFrontend,
)
from suppgram.frontends.telegram.app_manager import TelegramAppManager
from suppgram.frontends.telegram.callback_actions import CallbackActionKind
from suppgram.frontends.telegram.helpers import encode_callback_data, decode_callback_data
from suppgram.frontends.telegram.identification import (
    make_customer_identification,
    make_customer_diff,
)
from suppgram.frontends.telegram.storage import TelegramStorage, TelegramMessageKind
from suppgram.texts.interface import TextProvider

logger = logging.getLogger(__name__)


class TelegramCustomerFrontend(CustomerFrontend):
    """
    Allows customers to access the support system via Telegram bot.

    All messages from all users to a specific Telegram bot will be considered
    messages to the support. When agent responds, the bot will copy agent's
    message content and send it to the customer. Important: the message will be
    copied, not forwarded; thus the agent's identity will remain hidden.
    """

    def __init__(
        self,
        token: str,
        app_manager: TelegramAppManager,
        backend: Backend,
        storage: TelegramStorage,
        texts: TextProvider,
    ):
        """
        This constructor should not be used directly; use [Builder][suppgram.builder.Builder] instead.

        Arguments:
            token: Telegram bot token.
            app_manager: helper object storing built `telegram.ext.Application` instances.
            backend: used backend instance.
            storage: helper object encapsulating persistent storage of Telegram-specific data.
            texts: texts provider.
        """
        self._backend = backend
        self._texts = texts
        self._storage = storage
        self._telegram_app = app_manager.get_app(token)
        self._telegram_bot: Bot = self._telegram_app.bot
        self._telegram_app.add_handlers(self._make_handlers())
        self._backend.on_new_message_for_customer.add_handler(
            self._handle_new_message_for_customer_event
        )
        self._backend.on_conversation_rated.add_handler(self._handle_conversation_rated)

    def _make_handlers(self) -> List[BaseHandler]:
        return [
            CommandHandler("start", self._handle_start_command),
            CallbackQueryHandler(self._handle_callback_query),
            MessageHandler(ChatType.PRIVATE & (TEXT | PHOTO), self._handle_text_message),
            MessageHandler(ChatType.PRIVATE & ~(TEXT | PHOTO), self._handle_unsupported_message),
        ]

    async def initialize(self):
        await super().initialize()
        await self._telegram_app.initialize()

    async def start(self):
        await self._telegram_app.updater.start_polling()
        await self._telegram_app.start()

    async def _handle_start_command(self, update: Update, context: ContextTypes.DEFAULT_TYPE):
        assert update.effective_chat, "command update should have `effective_chat`"
        await context.bot.send_message(
            chat_id=update.effective_chat.id, text=self._texts.telegram_customer_start_message
        )

    async def _handle_callback_query(self, update: Update, context: ContextTypes.DEFAULT_TYPE):
        assert update.callback_query, "callback query update should have `callback_query`"
        assert update.effective_chat, "callback query update should have `effective_chat`"
        assert update.effective_user, "callback query update should have `effective_user`"
        if not update.callback_query.data:
            # No idea how to handle this update.
            return
        callback_data = decode_callback_data(update.callback_query.data)
        action = callback_data["a"]
        conversation = await self._backend.identify_customer_conversation(
            make_customer_identification(update.effective_user)
        )
        conversation_to_rate = await self._backend.get_conversation(callback_data["c"])
        if action == CallbackActionKind.RATE:
            if conversation.customer.id != conversation_to_rate.customer.id:
                # Can't rate another person's conversation!
                # Probably someone's playing with the callback query payloads.
                return
            await self._backend.rate_conversation(conversation_to_rate, callback_data["r"])
        else:
            logger.info(f"Customer frontend received unsupported callback action {action!r}")

    async def _handle_text_message(self, update: Update, context: ContextTypes.DEFAULT_TYPE):
        assert update.message, "message update should have `message`"
        assert update.effective_user, "message update should have `effective_user`"
        identification = make_customer_identification(update.effective_user)
        await self._backend.create_or_update_customer(
            identification,
            make_customer_diff(update.effective_user),
        )
        conversation = await self._backend.identify_customer_conversation(identification)
        if update.message.photo:
            photo = max(update.message.photo, key=lambda p: p.width * p.height)
            file = await context.bot.get_file(photo.file_id)
            data = await file.download_as_bytearray()
            await self._backend.process_message(
                conversation,
                Message(
                    kind=MessageKind.FROM_CUSTOMER,
                    time_utc=update.message.date,
                    image=bytes(data),
                ),
            )
            if update.message.caption:
                await self._backend.process_message(
                    conversation,
                    Message(
                        kind=MessageKind.FROM_CUSTOMER,
                        time_utc=update.message.date,
                        text=update.message.caption,
                    ),
                )
        if update.message.text:
            await self._backend.process_message(
                conversation,
                Message(
                    kind=MessageKind.FROM_CUSTOMER,
                    time_utc=update.message.date,
                    text=update.message.text,
                ),
            )
        # TODO batch processing of multiple messages

    async def _handle_unsupported_message(self, update: Update, context: ContextTypes.DEFAULT_TYPE):
        assert update.message, "message update should have `message`"
        assert update.effective_chat, "message update should have `effective_chat`"
        await self._telegram_bot.send_message(
            chat_id=update.effective_chat.id,
            text=self._texts.telegram_customer_unsupported_message_content,
            reply_to_message_id=update.message.message_id,
        )

    async def _handle_new_message_for_customer_event(self, event: NewMessageForCustomerEvent):
        customer = event.conversation.customer
        if not customer.telegram_user_id:
            return

        if event.message.kind == MessageKind.RESOLVED:
            await self._handle_conversation_resolution(event)
            return

        if event.message.text:
            await self._telegram_bot.send_message(
                chat_id=customer.telegram_user_id,
                text=event.message.text,
            )

        if event.message.image:
            await self._telegram_bot.send_photo(
                chat_id=customer.telegram_user_id,
                photo=event.message.image,
            )

    async def _handle_conversation_resolution(self, event: NewMessageForCustomerEvent):
        customer = event.conversation.customer
        if customer.telegram_user_id is None:
            return

        reply_markup = InlineKeyboardMarkup(
            [
                [
                    self._make_rate_button(event.conversation, 1),
                    self._make_rate_button(event.conversation, 2),
                    self._make_rate_button(event.conversation, 3),
                ],
                [
                    self._make_rate_button(event.conversation, 4),
                    self._make_rate_button(event.conversation, 5),
                ],
            ]
        )
        message = await self._telegram_bot.send_message(
            chat_id=customer.telegram_user_id,
            text=self._texts.telegram_customer_conversation_resolved_message_placeholder,
        )
        group = await self._storage.create_or_update_chat(customer.telegram_user_id)
        await self._storage.insert_message(
            self._telegram_bot.id,
            group,
            message.message_id,
            TelegramMessageKind.RATE_CONVERSATION,
            conversation_id=event.conversation.id,
        )
        await self._telegram_bot.edit_message_text(
            chat_id=customer.telegram_user_id,
            message_id=message.message_id,
            text=self._texts.telegram_customer_conversation_resolved_message,
            reply_markup=reply_markup,
        )

    async def _handle_conversation_rated(self, event: ConversationEvent):
        messages = await self._storage.get_messages(
            TelegramMessageKind.RATE_CONVERSATION, conversation_id=event.conversation.id
        )
        for message in messages:
            # Normally just one message, so no `gather()`.
            await self._telegram_bot.edit_message_text(
                chat_id=message.chat.telegram_chat_id,
                message_id=message.telegram_message_id,
                text=self._texts.compose_customer_conversation_resolved_message(
                    cast(int, event.conversation.customer_rating)
                ),
                reply_markup=None,
            )

    def _make_rate_button(self, conversation: Conversation, rating: int) -> InlineKeyboardButton:
        return InlineKeyboardButton(
            text=self._texts.format_rating(rating),
            callback_data=encode_callback_data(
                {"a": CallbackActionKind.RATE, "c": conversation.id, "r": rating}
            ),
        )



import asyncio
import logging
import os
import sys
from typing import Optional

import click
from click import UsageError

from suppgram.builder import Builder
from suppgram.errors import NoStorageSpecified, NoFrontendSpecified
from suppgram.logging import ConfidentialStreamHandler


@click.command()
@click.option(
    "--loglevel",
    type=click.Choice(["DEBUG", "INFO", "WARN", "WARNING", "ERROR", "FATAL", "CRITICAL"]),
    default="INFO",
    help="Log level",
)
@click.option(
    "--sqlalchemy-uri",
    envvar="SQLALCHEMY_URI",
    default=None,
    help="SQLAlchemy connection URI. Alternatively, environment variable SQLALCHEMY_URI may be used",
)
@click.option(
    "--sqlalchemy-uri-file",
    default=None,
    type=click.Path(exists=True, dir_okay=False),
    help="Path to file storing SQLAlchemy connection URI. "
    "Alternatively, environment variable SQLALCHEMY_URI may be used",
)
@click.option(
    "--mongodb-uri",
    envvar="MONGODB_URI",
    default=None,
    help="MongoDB connection URI. Alternatively, environment variable MONGODB_URI may be used",
)
@click.option(
    "--mongodb-uri-file",
    default=None,
    type=click.Path(exists=True, dir_okay=False),
    help="Path to file storing MongoDB connection URI. Alternatively, environment variable MONGODB_URI may be used",
)
@click.option(
    "--mongodb-database",
    "mongodb_database_name",
    default=None,
    help="MongoDB database name. If not specified, will connect to the default database specified in the URI",
)
@click.option(
    "--texts",
    "texts_class_path",
    default="suppgram.texts.en.EnglishTextProvider",
    show_default=True,
    help="Class with texts",
)
@click.option(
    "--telegram-customer-bot-token-file",
    default=None,
    type=click.Path(exists=True, dir_okay=False),
    help="Path to file storing token for Telegram bot serving customers. "
    "Alternatively, environment variable TELEGRAM_CUSTOMER_BOT_TOKEN may be used",
)
@click.option(
    "--telegram-manager-bot-token-file",
    default=None,
    type=click.Path(exists=True, dir_okay=False),
    help="Path to file storing token for Telegram manager bot. "
    "Alternatively, environment variable TELEGRAM_MANAGER_BOT_TOKEN may be used",
)
@click.option(
    "--telegram-agent-bot-tokens-file",
    default=None,
    type=click.Path(exists=True, dir_okay=False),
    help="Path to file storing token(s) for Telegram bot(s) serving agents. "
    "Alternatively, environment variable TELEGRAM_AGENT_BOT_TOKENS may be used",
)
@click.option(
    "--telegram-owner-id",
    type=int,
    default=None,
    help="ID of Telegram user who will be granted all permissions",
)
@click.option(
    "--customer-shell",
    is_flag=True,
    default=False,
    help="Run shell-based customer interface",
)
@click.option(
    "--pubnub-user-id",
    default="support",
    show_default=True,
    help="PubNub user ID to send support messages from",
)
@click.option(
    "--pubnub-channel-group",
    default="support",
    show_default=True,
    help="PubNub channel group containing chats with support",
)
@click.option(
    "--pubnub-message-converter",
    "pubnub_message_converter_class_path",
    default="suppgram.frontends.pubnub.DefaultMessageConverter",
    show_default=True,
    help="Class converting messages between PubNub JSONs and suppgram.entities.Message objects",
)
def run_all_in_one(
    loglevel: str,
    sqlalchemy_uri: Optional[str],
    sqlalchemy_uri_file: Optional[str],
    mongodb_uri: Optional[str],
    mongodb_uri_file: Optional[str],
    mongodb_database_name: Optional[str],
    texts_class_path: str,
    telegram_customer_bot_token_file: Optional[str],
    telegram_manager_bot_token_file: Optional[str],
    telegram_agent_bot_tokens_file: Optional[str],
    telegram_owner_id: Optional[int],
    customer_shell: bool,
    pubnub_user_id: str,
    pubnub_channel_group: str,
    pubnub_message_converter_class_path: str,
):
    telegram_customer_bot_token: Optional[str] = os.environ.get(
        "TELEGRAM_CUSTOMER_BOT_TOKEN"
    ) or _read_secret_from_file(telegram_customer_bot_token_file)

    telegram_manager_bot_token: Optional[str] = os.environ.get(
        "TELEGRAM_MANAGER_BOT_TOKEN"
    ) or _read_secret_from_file(telegram_manager_bot_token_file)

    telegram_agent_bot_tokens_joined: Optional[str] = os.environ.get(
        "TELEGRAM_AGENT_BOT_TOKENS"
    ) or _read_secret_from_file(telegram_agent_bot_tokens_file)
    telegram_agent_bot_tokens = (
        telegram_agent_bot_tokens_joined.split() if telegram_agent_bot_tokens_joined else []
    )

    replacements = {}
    if telegram_customer_bot_token:
        replacements[telegram_customer_bot_token] = "__CUSTOMER_BOT_TOKEN__"
    if telegram_manager_bot_token:
        replacements[telegram_manager_bot_token] = "__MANAGER_BOT_TOKEN__"
    for i, token in enumerate(telegram_agent_bot_tokens):
        replacements[token] = f"__AGENT_BOT_TOKEN_{i}__"
    logging.basicConfig(
        level=getattr(logging, loglevel),
        handlers=[ConfidentialStreamHandler(sys.stderr, replacements)],
    )

    builder = Builder()

    if sqlalchemy_uri_file:
        if sqlalchemy_uri:
            raise UsageError(
                "both --sqlalchemy-uri (or SQLALCHEMY_URI environment variable) "
                "and --sqlalchemy-uri-file are specified; unclear which URI to use."
            )

        sqlalchemy_uri = _read_secret_from_file(sqlalchemy_uri_file)

    try:
        if sqlalchemy_uri:
            builder = builder.with_sqlalchemy_storage(sqlalchemy_uri)
    except ImportError as exc:
        raise UsageError(
            "can't import SQLAlchemy integration. "
            "Make sure that necessary dependencies are installed:\n\n    pip install suppgram[sqlalchemy]\n"
        ) from exc

    if mongodb_uri_file:
        if mongodb_uri:
            raise UsageError(
                "both --mongodb-uri (or MONGODB_URI environment variable) "
                "and --mongodb-uri-file are specified; unclear which URI to use."
            )

        mongodb_uri = _read_secret_from_file(mongodb_uri_file)

    try:
        if mongodb_uri is not None:
            builder = builder.with_mongodb_storage(mongodb_uri, mongodb_database_name)
    except ImportError as exc:
        raise UsageError(
            "can't import MongoDB integration. "
            "Make sure that necessary dependencies are installed:\n\n    pip install suppgram[mongodb]\n"
        ) from exc

    if texts_class_path:
        builder = builder.with_texts_class_path(texts_class_path)

    try:
        if telegram_manager_bot_token:
            builder = builder.with_telegram_manager_frontend(
                telegram_manager_bot_token, telegram_owner_id
            )

        if telegram_customer_bot_token:
            builder = builder.with_telegram_customer_frontend(telegram_customer_bot_token)

        if telegram_agent_bot_tokens:
            builder = builder.with_telegram_agent_frontend(telegram_agent_bot_tokens)
    except ImportError as exc:
        raise UsageError(
            "can't import Telegram integration. "
            "Make sure that necessary dependencies are installed:\n\n    pip install suppgram[telegram]\n"
        ) from exc

    if customer_shell:
        builder = builder.with_shell_customer_frontend()

    if (
        any(arg.startswith("--pubnub") for arg in sys.argv)
        or "PUBNUB_SUBSCRIBE_KEY" in os.environ
        or "PUBNUB_PUBLISH_KEY" in os.environ
        or "PUBNUB_SECRET_KEY" in os.environ
        or "PUBNUB_TOKEN" in os.environ
    ):
        try:
            from suppgram.frontends.pubnub.errors import MissingCredentials
        except ImportError:
            raise UsageError(
                "can't import PubNub integration. "
                "Make sure that necessary dependencies are installed:\n\n    pip install suppgram[pubnub]\n"
            )

        try:
            builder = builder.with_pubnub_customer_frontend(
                pubnub_user_id=pubnub_user_id,
                pubnub_channel_group=pubnub_channel_group,
                pubnub_message_converter_class_path=pubnub_message_converter_class_path,
            )
        except MissingCredentials as exc:
            raise UsageError(str(exc)) from exc

    try:
        builder.build()
    except NoStorageSpecified as exc:
        raise UsageError(
            "no storage specified. Consider specifying --sqlalchemy-uri or --mongodb-uri parameters."
        ) from exc
    except NoFrontendSpecified as exc:
        raise UsageError(
            "no frontend specified. In this configuration the application is not going to do anything.\n"
            "Consider specifying --telegram-*, --pubnub-* or --customer-shell parameters."
        ) from exc

    loop = asyncio.get_event_loop()
    loop.run_until_complete(builder.start())
    loop.run_forever()


def _read_secret_from_file(path: Optional[str]) -> Optional[str]:
    if path is None:
        return None
    with open(path) as f:
        return f.read().strip()


if __name__ == "__main__":
    run_all_in_one()



from datetime import timezone
from typing import List, Optional, Any, Iterable, Mapping, Dict

from bson import CodecOptions, ObjectId
from motor.core import AgnosticDatabase
from pymongo import ReturnDocument

from suppgram.frontends.telegram.storage import (
    TelegramStorage,
    TelegramMessage,
    TelegramMessageKind,
    TelegramChat,
    TelegramChatRole,
)
from suppgram.storages.mongodb.collections import Document


class MongoDBTelegramBridge(TelegramStorage):
    """Implementation of [Storage][suppgram.storage.Storage] for MongoDB
    via [Motor](https://motor.readthedocs.io/) library."""

    def __init__(
        self,
        database: AgnosticDatabase,
        chat_collection_name: str = "suppgram_telegram_chats",
        messages_collection_name: str = "suppgram_telegram_messages",
    ):
        """
        Parameters:
            database: asyncio-compatible Motor MongoDB database to store data in
            chat_collection_name: name of MongoDB collection to store data on Telegram chats in
            messages_collection_name: name of MongoDB collection to store references to Telegram messages in
        """
        codec_options: CodecOptions = CodecOptions(tz_aware=True, tzinfo=timezone.utc)
        self._chat_collection = database.get_collection(chat_collection_name, codec_options)
        self._message_collection = database.get_collection(messages_collection_name, codec_options)

    async def get_chat(self, telegram_chat_id: int) -> TelegramChat:
        filter_ = self._make_chat_filter(telegram_chat_id)
        doc = await self._chat_collection.find_one(filter_)
        if doc is None:
            raise ValueError(f"couldn't find Telegram chat {telegram_chat_id}")
        return self._convert_to_chat(doc)

    async def find_chats_by_ids(self, telegram_chat_ids: List[int]) -> List[TelegramChat]:
        filter_ = self._make_chats_filter(telegram_chat_ids)
        docs = await self._chat_collection.find(filter_).to_list(None)
        return [self._convert_to_chat(doc) for doc in docs]

    async def create_or_update_chat(self, telegram_chat_id: int) -> TelegramChat:
        filter_ = self._make_chat_filter(telegram_chat_id)
        update = self._make_chat_update()
        doc = await self._chat_collection.find_one_and_update(
            filter_, update, upsert=True, return_document=ReturnDocument.AFTER
        )
        return self._convert_to_chat(doc)

    async def add_chat_roles(self, telegram_chat_id: int, *roles: TelegramChatRole):
        filter_ = self._make_chat_filter(telegram_chat_id)
        update = self._make_chat_roles_update(roles)
        result = await self._chat_collection.update_one(filter_, update)
        if result.matched_count == 0:
            raise ValueError(f"couldn't find Telegram chat {telegram_chat_id}")

    async def get_chats_by_role(self, role: TelegramChatRole) -> List[TelegramChat]:
        filter_ = self._make_chat_filter_by_role(role)
        docs = await self._chat_collection.find(filter_).to_list(None)
        return [self._convert_to_chat(doc) for doc in docs]

    def _make_chat_filter(self, telegram_chat_id: int) -> Document:
        return {"_id": telegram_chat_id}

    def _make_chats_filter(self, telegram_chat_ids: List[int]) -> Document:
        return {"_id": {"$in": telegram_chat_ids}}

    def _make_chat_filter_by_role(self, role: TelegramChatRole) -> Document:
        return {"roles": role.value}

    def _make_chat_update(self) -> Document:
        return {"$setOnInsert": {"roles": []}}

    def _make_chat_roles_update(self, roles: Iterable[TelegramChatRole]) -> Document:
        return {"$addToSet": {"roles": {"$each": [role.value for role in roles]}}}

    def _convert_to_chat(self, doc: Document) -> TelegramChat:
        return TelegramChat(
            telegram_chat_id=doc["_id"],
            roles=frozenset(TelegramChatRole(role) for role in doc.get("roles", [])),
        )

    async def insert_message(
        self,
        telegram_bot_id: int,
        chat: TelegramChat,
        telegram_message_id: int,
        kind: TelegramMessageKind,
        *,
        agent_id: Optional[Any] = None,
        customer_id: Optional[Any] = None,
        conversation_id: Optional[Any] = None,
        telegram_bot_username: Optional[str] = None,
    ) -> TelegramMessage:
        doc = {
            "_id": self._compose_message_id(chat.telegram_chat_id, telegram_message_id),
            "telegram_bot_id": telegram_bot_id,
            "kind": kind,
        }
        if agent_id is not None:
            doc["agent_id"] = ObjectId(agent_id)
        if customer_id is not None:
            doc["customer_id"] = ObjectId(customer_id)
        if conversation_id is not None:
            doc["conversation_id"] = ObjectId(conversation_id)
        if telegram_bot_username is not None:
            doc["telegram_bot_username"] = telegram_bot_username
        await self._message_collection.insert_one(doc)
        return self._convert_to_message(doc, {chat.telegram_chat_id: chat})

    async def get_message(self, chat: TelegramChat, telegram_message_id: int) -> TelegramMessage:
        filter_ = {"_id": self._compose_message_id(chat.telegram_chat_id, telegram_message_id)}
        doc = await self._message_collection.find_one(filter_)
        if doc is None:
            raise ValueError(
                f"couldn't find Telegram message {telegram_message_id} in chat {chat.telegram_chat_id}"
            )
        return self._convert_to_message(doc, chats={chat.telegram_chat_id: chat})

    async def get_messages(
        self,
        kind: TelegramMessageKind,
        *,
        agent_id: Optional[Any] = None,
        conversation_id: Optional[Any] = None,
        telegram_bot_username: Optional[str] = None,
    ) -> List[TelegramMessage]:
        filter_: Dict[str, Any] = {"kind": kind}
        if agent_id is not None:
            filter_["agent_id"] = ObjectId(agent_id)
        if conversation_id is not None:
            filter_["conversation_id"] = ObjectId(conversation_id)
        if telegram_bot_username is not None:
            filter_["telegram_bot_username"] = telegram_bot_username
        docs = await self._message_collection.find(filter_).to_list(None)
        return await self._convert_to_messages(docs)

    async def delete_messages(self, messages: List[TelegramMessage]):
        message_ids = [
            self._compose_message_id(message.chat.telegram_chat_id, message.telegram_message_id)
            for message in messages
        ]
        filter_ = {"_id": {"$in": message_ids}}
        await self._message_collection.delete_many(filter_)

    async def get_newer_messages_of_kind(
        self, messages: List[TelegramMessage]
    ) -> List[TelegramMessage]:
        if not messages:
            return []
        filters_per_chat = [
            {
                "_id.c": message.chat.telegram_chat_id,
                "_id.m": {"$gt": message.telegram_message_id},
                "kind": message.kind,
            }
            for message in messages
        ]
        filter_ = {"$or": filters_per_chat}
        docs = await self._message_collection.find(filter_).to_list(None)
        return await self._convert_to_messages(docs)

    async def _convert_to_messages(self, docs: List[Document]) -> List[TelegramMessage]:
        chat_ids = [doc["_id"]["c"] for doc in docs]
        chats = {chat.telegram_chat_id: chat for chat in await self.find_chats_by_ids(chat_ids)}
        return [self._convert_to_message(doc, chats) for doc in docs]

    def _compose_message_id(self, telegram_chat_id: int, telegram_message_id: int) -> Any:
        return {"c": telegram_chat_id, "m": telegram_message_id}

    def _convert_to_message(
        self, doc: Document, chats: Mapping[int, TelegramChat]
    ) -> TelegramMessage:
        telegram_chat_id = doc["_id"]["c"]
        telegram_message_id = doc["_id"]["m"]
        return TelegramMessage(
            id=f"{telegram_chat_id}_{telegram_message_id}",
            telegram_bot_id=doc["telegram_bot_id"],
            chat=chats[telegram_chat_id],
            telegram_message_id=telegram_message_id,
            kind=TelegramMessageKind(doc["kind"]),
            agent_id=str(doc["agent_id"]) if "agent_id" in doc else None,
            customer_id=str(doc["customer_id"]) if "customer_id" in doc else None,
            conversation_id=str(doc["conversation_id"]) if "conversation_id" in doc else None,
            telegram_bot_username=doc.get("telegram_bot_username"),
        )



import operator
from functools import reduce
from typing import Optional, Any, List

from sqlalchemy import (
    Integer,
    ForeignKey,
    Enum,
    select,
    update,
    ColumnElement,
    String,
    delete,
)
from sqlalchemy.ext.asyncio import AsyncEngine, async_sessionmaker
from sqlalchemy.orm import Mapped, mapped_column, relationship, joinedload

from suppgram.frontends.telegram.storage import (
    TelegramStorage,
    TelegramChat as TelegramGroupInterface,
    TelegramMessage as TelegramMessageInterface,
    TelegramMessageKind,
    TelegramChatRole,
)
from suppgram.storages.sqlalchemy.models import Base, Conversation, Customer, Agent


class TelegramChat(Base):
    __tablename__ = "suppgram_telegram_chats"
    telegram_chat_id: Mapped[int] = mapped_column(Integer, primary_key=True)
    roles: Mapped[int] = mapped_column(Integer, default=0)
    messages: Mapped[List["TelegramMessage"]] = relationship(back_populates="chat")


class TelegramMessage(Base):
    __tablename__ = "suppgram_telegram_messages"
    id: Mapped[int] = mapped_column(Integer, primary_key=True, autoincrement=True)
    chat_id: Mapped[int] = mapped_column(ForeignKey(TelegramChat.telegram_chat_id), nullable=False)
    telegram_bot_id: Mapped[int] = mapped_column(Integer, nullable=False)
    chat: Mapped[TelegramChat] = relationship(back_populates="messages")
    telegram_message_id: Mapped[int] = mapped_column(Integer, nullable=False)
    kind: Mapped[TelegramMessageKind] = mapped_column(Enum(TelegramMessageKind), nullable=False)
    agent_id: Mapped[int] = mapped_column(ForeignKey(Agent.id), nullable=True)
    customer_id: Mapped[int] = mapped_column(ForeignKey(Customer.id), nullable=True)
    conversation_id: Mapped[int] = mapped_column(ForeignKey(Conversation.id), nullable=True)
    telegram_bot_username: Mapped[str] = mapped_column(String, nullable=True)


class SQLAlchemyTelegramBridge(TelegramStorage):
    """Implementation of [TelegramStorage][suppgram.frontends.telegram.TelegramStorage] for SQLAlchemy."""

    def __init__(
        self,
        engine: AsyncEngine,
        chat_model: Any = TelegramChat,
        message_model: Any = TelegramMessage,
    ):
        """
        Parameters:
            engine: asynchronous SQLAlchemy engine
            chat_model: SQLAlchemy model for Telegram chats
            message_model: SQLAlchemy model for Telegram messages
        """
        self._engine = engine
        self._session = async_sessionmaker(bind=engine)
        self._chat_model = chat_model
        self._message_model = message_model

    async def initialize(self):
        await super().initialize()
        tables_to_create = [
            built_in_model.__table__
            for model, built_in_model in [
                (self._chat_model, TelegramChat),
                (self._message_model, TelegramMessage),
            ]
            if model is built_in_model
        ]
        async with self._engine.begin() as conn:
            await conn.run_sync(Base.metadata.create_all, tables=tables_to_create)

    async def get_chats_by_role(self, role: TelegramChatRole) -> List[TelegramGroupInterface]:
        select_query = select(self._chat_model).filter(
            self._chat_model.roles.bitwise_and(role.value) != 0
        )
        async with self._session() as session:
            chats = (await session.execute(select_query)).scalars().all()
            return [self._convert_chat(chat) for chat in chats]

    async def get_chat(self, telegram_chat_id: int) -> TelegramGroupInterface:
        select_query = select(self._chat_model).filter(
            self._chat_model.telegram_chat_id == telegram_chat_id
        )
        async with self._session() as session:
            chat = (await session.execute(select_query)).scalars().one()
            return self._convert_chat(chat)

    async def create_or_update_chat(self, telegram_chat_id: int) -> TelegramGroupInterface:
        async with self._session() as session, session.begin():
            query = select(self._chat_model).filter(
                self._chat_model.telegram_chat_id == telegram_chat_id
            )
            chat = (await session.execute(query)).scalars().one_or_none()
            if chat is None:
                chat = self._chat_model(telegram_chat_id=telegram_chat_id)  # type: ignore
                session.add(chat)
            return self._convert_chat(chat)

    async def add_chat_roles(self, telegram_chat_id: int, *roles: TelegramChatRole):
        role_values_or = reduce(operator.or_, (role.value for role in roles))
        update_query = (
            update(TelegramChat)
            .filter(TelegramChat.telegram_chat_id == telegram_chat_id)
            .values(roles=TelegramChat.roles.bitwise_or(role_values_or))
        )
        async with self._session() as session, session.begin():
            await session.execute(update_query)

    def _convert_chat(self, chat: TelegramChat) -> TelegramGroupInterface:
        roles: List[TelegramChatRole] = []
        role, encoded_roles = 1, chat.roles
        while encoded_roles:
            if role & encoded_roles:
                roles.append(TelegramChatRole(role))
                encoded_roles ^= role
            role = role << 1
        return TelegramGroupInterface(
            telegram_chat_id=chat.telegram_chat_id, roles=frozenset(roles)
        )

    async def insert_message(
        self,
        telegram_bot_id: int,
        chat: TelegramGroupInterface,
        telegram_message_id: int,
        kind: TelegramMessageKind,
        *,
        agent_id: Optional[Any] = None,
        customer_id: Optional[Any] = None,
        conversation_id: Optional[Any] = None,
        telegram_bot_username: Optional[str] = None,
    ) -> TelegramMessageInterface:
        async with self._session() as session, session.begin():
            message = self._message_model(
                telegram_bot_id=telegram_bot_id,
                chat_id=chat.telegram_chat_id,
                telegram_message_id=telegram_message_id,
                kind=kind,
                agent_id=agent_id,
                customer_id=customer_id,
                conversation_id=conversation_id,
                telegram_bot_username=telegram_bot_username,
            )
            session.add(message)
            await session.flush()
            await session.refresh(message)
            return self._convert_message(message, chat)

    async def get_message(
        self, chat: TelegramGroupInterface, telegram_message_id: int
    ) -> TelegramMessageInterface:
        filter_ = (self._message_model.chat_id == chat.telegram_chat_id) & (
            self._message_model.telegram_message_id == telegram_message_id
        )
        select_query = (
            select(self._message_model).options(joinedload(self._message_model.chat)).where(filter_)
        )
        async with self._session() as session:
            msg = (await session.execute(select_query)).scalars().one()
            return self._convert_message(msg, self._convert_chat(msg.chat))

    async def get_messages(
        self,
        kind: TelegramMessageKind,
        *,
        agent_id: Optional[Any] = None,
        conversation_id: Optional[Any] = None,
        telegram_bot_username: Optional[str] = None,
    ) -> List[TelegramMessageInterface]:
        filter_ = self._message_model.kind == kind
        if agent_id is not None:
            filter_ = filter_ & (self._message_model.agent_id == agent_id)
        if conversation_id is not None:
            filter_ = filter_ & (self._message_model.conversation_id == conversation_id)
        if telegram_bot_username:
            filter_ = filter_ & (self._message_model.telegram_bot_username == telegram_bot_username)
        select_query = (
            select(self._message_model).options(joinedload(self._message_model.chat)).where(filter_)
        )
        async with self._session() as session:
            msgs = (await session.execute(select_query)).scalars().all()
            return [self._convert_message(msg, self._convert_chat(msg.chat)) for msg in msgs]

    async def delete_messages(self, messages: List[TelegramMessageInterface]):
        message_ids = [message.id for message in messages]
        delete_query = delete(self._message_model).where(self._message_model.id.in_(message_ids))
        async with self._session() as session, session.begin():
            await session.execute(delete_query)

    async def get_newer_messages_of_kind(
        self, messages: List[TelegramMessageInterface]
    ) -> List[TelegramMessageInterface]:
        if not messages:
            return []
        async with self._session() as session:
            filter = reduce(
                operator.or_,
                (self._filter_newer_messages_of_kind(message) for message in messages),
            )
            select_query = (
                select(self._message_model)
                .options(joinedload(self._message_model.chat))
                .filter(filter)
            )
            msgs = (await session.execute(select_query)).scalars().all()
            return [self._convert_message(msg, self._convert_chat(msg.chat)) for msg in msgs]

    def _filter_newer_messages_of_kind(self, message: TelegramMessageInterface) -> ColumnElement:
        return (
            (self._message_model.kind == message.kind)
            & (self._message_model.chat_id == message.chat.telegram_chat_id)
            & (self._message_model.telegram_message_id > message.telegram_message_id)
        )

    def _convert_message(
        self, msg: TelegramMessage, chat: TelegramGroupInterface
    ) -> TelegramMessageInterface:
        return TelegramMessageInterface(
            id=msg.id,
            telegram_bot_id=msg.telegram_bot_id,
            chat=chat,
            telegram_message_id=msg.telegram_message_id,
            kind=msg.kind,
            agent_id=msg.agent_id,
            customer_id=msg.customer_id,
            conversation_id=msg.conversation_id,
            telegram_bot_username=msg.telegram_bot_username,
        )
